import json
import logging
import numbers
import os
import pexpect
import socket
import urllib.request, urllib.error, urllib.parse

import jsonrpclib

from pexpect import popen_spawn
from decimal import Decimal

from .processing import inexact_to_decimal, wait_for_success, check_port


logger = logging.getLogger('piggy_logs')


class PiggyBTC:
    def __init__(
            self,
            wallet_bin_path,
            datastore_path,
            wallet_filename,
            wallet_password,
            rpcuser,
            rpcpassword,
            rpcport
            ):

        """Manage a Bitcoin wallet using Electrum.

        Supported wallet version:
        https://download.electrum.org/3.1.3/Electrum-3.1.3.tar.gz

        :param wallet_bin_path: Path to Electrum wallet executable
        :param datastore_path: Path to datastore directory (which includes wallet files and block headers)
        :param wallet_filename: Name of wallet file to use (must be in `datastore_path/wallets`)
        :param wallet_password: Password to enter for decrypting wallet
        :param rpcport: Local port to start the wallet RPC server on

        """

        self.wallet_bin_path = wallet_bin_path
        self.datadir = datastore_path
        self.wallet_filename = wallet_filename
        self.wallet_password = wallet_password

        self.rpcuser = rpcuser
        self.rpcpassword = rpcpassword

        self.rpcport = check_port(rpcport, 'rpcport')
        self.server = None

    def _execute_electrum_command(
            self,
            command,
            expect=None,
            stdin_text=None,
            quiet=False,
            wait_for_command=True
            ):

        """
        Runs commands using the Electrum binary

        :param command: List of string CLI arguments for the Electrum wallet
        :param expect: Command output to expect, wait for, read, and match.
        :param stdin_text: String or None; if string: pass in string as stdin to Electrum
        :param quiet: Suppress printing process details
        :param wait_for_command: Wait for the program to end (blocking call).

            If true, it uses pexpect with  a pseudo-terminal, but the terminal quits in the following cases:
                - when the command quits (and it kills the command's children processes)
                - when the Python script is interrupted
            If false, it uses Popen, which does not provide a pseudo-terminal (losing, say, password input
            capability), but it lets the spawned process and its children survive past the Python script.
        """

        done_str = 'Check stdout/stderr here.' if stdin_text else ''

        if quiet is False:
            logger.info('#######################')
            logger.info('Executing "{}". {}'.format(command, done_str))

        wallet_file = os.path.join(self.datadir, 'wallets', self.wallet_filename)

        self._check_paths(wallet_file)

        if wait_for_command:
            process = pexpect.spawn(
                self.wallet_bin_path,
                command + [
                    '--dir', self.datadir,
                    '--wallet', wallet_file
                ]
            )

            if expect:
                process.expect_exact(expect)

            if stdin_text is not None:
                process.sendline(stdin_text)

            output = process.read()

            if quiet is False:
                logger.warning(
                    'Process printed EOF.\n'
                    '### OUTPUT:\n{}\n###'.format(output)
                )

            status = process.wait()

            if quiet is False:
                logger.info("Exit code {}".format(status))

        else:
            if expect or stdin_text:
                raise ValueError(
                    "Can not properly expect or send text if we don't wait for command."
                )

            process = popen_spawn.PopenSpawn(
                [self.wallet_bin_path] +
                command + [
                    '--dir', self.datadir,
                    '--wallet', wallet_file
                ]
            )

            output = None

        return process, output

    def _check_paths(self, wallet_file):
        """
        Ensure existence of wallet binary, datadir, and wallet file

        :param wallet_file: Path to wallet file
        """

        os.stat(os.path.abspath(self.wallet_bin_path))
        try:
            os.stat(os.path.abspath(self.datadir))
        except OSError:
            os.makedirs(os.path.abspath(self.datadir))

        try:
            os.stat(os.path.abspath(wallet_file))
        except OSError:
            logger.critical(
                'Missing wallet file "{}"!\n Please create the wallet or change the path.'
                .format(os.path.abspath(wallet_file))
            )
            raise

    def _connect_if_needed(self):
        if self.server is None:
            self.server = jsonrpclib.Server("http://{}:{}@127.0.0.1:{}".format(
                    self.rpcuser,
                    self.rpcpassword,
                    self.rpcport
            ))

    def _check_version(self):
        """Check that the version is supported."""

        try:
            _, status = self._execute_electrum_command(
                    ['version'],
                    expect='3.1.3',
                    quiet=True
            )
        except pexpect.exceptions.EOF:
            raise ValueError('Electrum help payto does not match our expectation.')

    def start_server(self):
        """Start the RPC server

        Use the wallet and datastore paths specified in config.
        """

        self._execute_electrum_command(['setconfig', 'rpcuser', str(self.rpcuser)])
        self._execute_electrum_command(['setconfig', 'rpcpassword', str(self.rpcpassword)])
        self._execute_electrum_command(['setconfig', 'rpcport', str(self.rpcport)])
        self._execute_electrum_command(['daemon', 'start'], wait_for_command=False)

        # Wait for daemon to start
        wait_for_success(self._daemon_rpc_loaded, 'BTC Daemon RPC')

        # Send password (this is NOT stdin)
        self._execute_electrum_command(
                ['daemon', 'load_wallet'],
                ['Password:', 'true'], self.wallet_password
                )

        self._check_version()

    def _daemon_rpc_loaded(self):
        """Check whether RPC server responds to a trivial request"""
        self._connect_if_needed()
        try:
            self.server.version()
            return True
        except socket.error:
            return False

    def stop_server(self):
        self._connect_if_needed()
        self._execute_electrum_command(['daemon', 'stop'])

    def get_balance(self):
        """Return the amount we have in the wallet, confirmed by the blockchain"""
        self._connect_if_needed()

        return inexact_to_decimal(self.server.getbalance()['confirmed'])

    def get_receive_address(self):
        """Returns an unused address of this wallet, where we can receive BTC

        Note: if you use an imported keystore instead of a seeded wallet,
        then `getunusedaddress()` may return None.
        """
        self._connect_if_needed()

        address = self.server.getunusedaddress()
        if not address:
            raise ValueError('Error: can not get unused address!')

        return address

    @staticmethod
    def suggest_miner_fee():
        """Returns miner fee to use, for typical transaction

        We should use fee estimates from Electrum, but it has no API for this.
        """
        request = urllib.request.Request(
            'https://bitcoinfees.21.co/api/v1/fees/recommended',
            headers={
                'User-Agent': 'Mozilla/5.0 (X11; CrOS i686 2268.111.0) AppleWebKit/536.11 (KHTML, like Gecko) Chrome/20.0.1132.57 Safari/536.11'
            }
        )
        response = urllib.request.urlopen(request)
        data = json.loads(response.read())

        #   The above is in Satoshis/B
        #   Multiply by 226 (typical bytes in a txn) -> Satoshis/txn
        #   Multiply by 0.00000001 (BTCs in a Sat) -> BTC/txn

        ten_min_fee_satoshi_per_byte = data['halfHourFee']
        ten_min_fee_satoshi_per_txn = ten_min_fee_satoshi_per_byte * 226
        fee = Decimal(ten_min_fee_satoshi_per_txn) * Decimal('1e-8')

        if fee > Decimal('0.001'):
            raise ValueError('Fee greater than 1 mBTC!')
        else:
            logger.warning('BTC Miner fee estimate: {}'.format(fee))

        return fee

    def transactions_since(self, since_unix_time):
        """Gets incoming transactions since specified time

        Lists all transactions of positive amount (which increased our balance)
        having the `blocktime` >= the given unix time.

        :param since_unix_time: integer or float, unix time
        :return transactions: List of transaction dictionaries
            Each dict has `amount`, `txid`, and `time` of each transaction
        """
        self._connect_if_needed()

        if not isinstance(since_unix_time, numbers.Real):
            raise ValueError('since_unix_time must be a number!')

        history = json.loads(self.server.history())['transactions']

        return self._process_history(history, since_unix_time)

    @classmethod
    def _process_history(cls, history, since_unix_time):
        """Process the transaction history data"""
        for tx in history:
            amount, currency = tx['value'].split(' ')

            if currency != 'BTC':
                raise ValueError('May not treat non-BTC currencies from PiggyBTC!')

            tx['value_decimal'] = Decimal(amount)

        recently_received = [
            tx for tx in history if
                (tx['timestamp'] >= since_unix_time) and
                (tx['confirmations'] >= 1) and
                (tx['value_decimal'] > 0)  # We received money (not sent, and not zero)
        ]

        # Only get the required fields, name them consistently, and cast value to Decimal
        recently_received = [
            {
                'txid': tx['txid'],
                'time': tx['timestamp'],
                'value': inexact_to_decimal(tx['value_decimal'])
            } for tx in recently_received
        ]

        return recently_received

    def perform_transaction(self, net_amount, miner_fee, target_address):
        """Send Bitcoins to target_address (total cost: net_amount + miner_fee)"""

        self._connect_if_needed()

        assert isinstance(net_amount, Decimal)
        assert isinstance(miner_fee, Decimal)
        self._validate_address(target_address)

        tx = self.server.payto(
            destination=str(target_address),
            amount=str(net_amount),
            fee=str(miner_fee),
            password=self.wallet_password
        )

        success, txid = self.server.broadcast(tx['hex'])

        # Signed TX: {u'hex': u'0100000001718276ce6c2298bf71352fa38fb9b4d3d40f49ec592d0dae640d9f4709e11a6a000000006b483045022100a1cbbd78b0f6f3c01bac6ea0590adbe2941b2be3d08820e1adb5b42777c4bd6202201c176a5f0a25c20f856aa91c09a7581d4d1057fa4e644c65ee34d9fd113291d6012103c64230ff95961bda955ae374e90cf4d9883432731a2df0389204162b7e6dc579feffffff0107003700000000001976a91490c235fba4b7ea93106c7439703779cba6db772e88ac00000000', u'complete': True, u'final': True}
        # Broadcast TX: [True, u'f0619d982344c12169e5fbf38f83b474be7df5a009351a6b1aeb4430566fa7ba']

        if success:
            logger.warning("Signed TX: {}".format(tx))
            logger.warning("Broadcast TX id: {}".format(txid))
            return txid
        else:
            raise BaseException("Transaction broadcast fail: {}".format(txid))

    def _validate_address(self, btc_address):
        """Raises an exception if address is invalid"""
        self.server.getaddressbalance(btc_address)



import logging
from decimal import Decimal

from .piggy_btc import PiggyBTC
from .piggy_xmr import PiggyXMR
from .piggy_eth import PiggyETH

logger = logging.getLogger('piggy_logs')

class MasterPiggy:
    """
    One Piggy to rule them all!
    This class is used to manage money in all the other piggies.

    :param piggy_settings: Dictionary to specify piggy options. Example:
        `{'BTC': {'wallet_path': '/path/btcwallet/', [...]}, [...]}
    """

    supported = {
            'BTC': PiggyBTC,
            'XMR': PiggyXMR,
            'ETH': PiggyETH,
    }

    def __init__(self, piggy_settings):
        self.currencies = piggy_settings.keys()
        self._check_currency_support()

        # Initialize the requested piggies
        self.piggies = {
            currency: self.supported[currency](
                **(piggy_settings[currency])
            )
            for currency in self.currencies
        }

    def _check_currency_support(self):
        """Check that all required piggies have been initialized"""

        required = set(self.currencies)
        supported = set(self.supported.keys())
        missing = required - supported

        if missing:
            raise ValueError(
                'MasterPiggy: Missing support for configured currencies: {}'.format(missing)
            )

    def start_servers(self):
        for p in self.currencies:
            logger.info("Starting RPC server for {}...".format(p))
            self.piggies[p].start_server()

    def stop_servers(self):
        for p in self.currencies:
            logger.info("Stopping RPC server for {}...".format(p))
            self.piggies[p].stop_server()

    def get_balances(self):
        return {p: self.piggies[p].get_balance() for p in self.currencies}

    def get_receive_address(self, currency):
        return self.piggies[currency].get_receive_address()

    def suggest_miner_fee(self, currency):
        return self.piggies[currency].suggest_miner_fee()

    def transactions_since(self, currency, earliest_possible_received=0):
        return self.piggies[currency].transactions_since(earliest_possible_received)

    def perform_transaction(
            self,
            currency,
            net_amount,
            miner_fee,
            target_address
    ):
        assert isinstance(net_amount, Decimal)
        assert isinstance(miner_fee, Decimal)

        self.piggies[currency].perform_transaction(net_amount, miner_fee, target_address)



import logging

from decimal import Decimal
from time import sleep

logger = logging.getLogger('piggy_logs')


def inexact_to_decimal(inexact_float):
    """
    Transforms a float to decimal, through its __str__() method.

    >>> inexact_to_decimal(3.14)
    Decimal('3.14')

    Avoids doing the following:

    >>> Decimal(3.14)
    Decimal('3.140000000000000124344978758017532527446746826171875')

    """

    return Decimal(str(inexact_float))


def wait_for_success(success_check, description, max_seconds=120):
    """
    Wait until calling `success_check` returns a truthy value.

    :param success_check: Callable returning truthy on success, false otherwise
    :param description: What the task is about (to log)
    :param max_seconds: Maximum amount of time to wait, before giving up
    """
    logger.info('Waiting for {} (at most {} seconds)'.format(description, max_seconds))
    for wait in range(max_seconds):
        if success_check():
            return
        else:
            logger.info('Waited for {} ({} seconds so far)'.format(description, wait))
            sleep(1)

    if not success_check():
        raise ValueError('{} failed after {} seconds'.format(description, max_seconds))


def check_port(port_num, name='port'):
    """Check that the port is a valid number. You'd be surprised."""

    if not isinstance(port_num, int):
        raise ValueError('Error! {} ({}) is not an integer!'.format(name, port_num))
    elif port_num > 65535:
        raise ValueError('Error! {} number too large ({}})'.format(name, port_num))
    else:
        return port_num



#!/usr/bin/env python3

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='piggies',
    version='0.0.1',
    description='A package to automatically manage cryptocurrency wallets',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/danuker/piggies',
    license='MIT',
    author='Dan Gheorghe Haiduc',
    author_email='danuthaiduc@gmail.com',
    python_requires='>=3.5',
    packages=['piggies'],
    install_requires=['jsonrpclib-pelix', 'web3', 'pyetherchain', 'pexpect'],
    test_suite='nose.collector',
    tests_require=['nose'],
    classifiers=(
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
    ),
)



# Command used by Ethereum Wallet for fast sync:
# /home/dan/.config/Ethereum Wallet/binaries/Geth/unpacked/geth --syncmode fast --cache 1024 --datadir /media/dan/Wilderness/Blockchains/eth/

import calendar
import datetime
import logging
import os
import pexpect
import web3

from decimal import Decimal
from pexpect import popen_spawn
from pyetherchain.pyetherchain import EtherChain
from web3 import Web3

from .processing import wait_for_success, check_port

logger = logging.getLogger('piggy_logs')

class PiggyETH:
    TXN_GAS_LIMIT = 21000

    def __init__(self,
                  wallet_bin_path,
                  datastore_path,
                  wallet_password,
                  ):
        """
        Manage a Parity wallet for Ethereum.

        The wallet key file must be in `datastore_path/keys/ethereum/`.
	Multiple "accounts" (wallet key files) are not supported - see the `accounts` property.

        :param wallet_bin_path: Path to Parity wallet executable
        :param datastore_path: Path to datastore directory (which includes wallet files and blockchain)
        :param wallet_password: Password to enter for decrypting the account
        """

        self.wallet_bin_path = wallet_bin_path
        self.datastore_path = datastore_path
        self.wallet_password = wallet_password

        self._server = None
        self._accounts = None
        self._block_timestamps = {}
        self.ec = EtherChain()

    @property
    def server(self):
        if self._server is None:
            ipc = Web3.IPCProvider(
                os.path.join(self.datastore_path, 'jsonrpc.ipc')
            )

            self._server = Web3(ipc)
        return self._server

    @property
    def accounts(self):
        if self._accounts is None:
            self._accounts = self.server.eth.accounts

            # We only handle 1 account (address) as of now.
            # In order to support multiple accounts you need to alter:
            # `get_receive_address`, `get_balance`, `transactions_since`, `perform_transaction`
            assert len(self.accounts) == 1
        return self._accounts

    def start_server(self):
        if self._rpc_loaded():
            return

        command = [
            self.wallet_bin_path,
            '-d', self.datastore_path,
            '--log-file', os.path.join(os.getcwd(), self.datastore_path, 'parity_log.txt'),
            '--no-ancient-blocks',
            '--no-ws',
            '--no-jsonrpc',
            '--ipc-apis=web3,eth,personal', # We need personal to actually perform a transaction
            '--warp-barrier', '5842205',
        ]

        logger.info("Starting ETH wallet: {}".format(' '.join(command)))
        popen_spawn.PopenSpawn(command)

        wait_for_success(self._rpc_loaded, 'ETH RPC')

    def stop_server(self):
        """Use pkill to kill the parity wallet."""
        p = pexpect.spawn('/usr/bin/pkill', ['-f', self.wallet_bin_path])
        p.wait()
        if p.status is not 0:
            raise ValueError('Error pkilling ETH:\n{}'.format(p.read()))

    def _rpc_loaded(self):
        """Attempt a connection to the RPC"""
        try:
            self.server.eth.getBlock(self.server.eth.blockNumber)
            return True
        except (web3.utils.threads.Timeout, ConnectionRefusedError, FileNotFoundError):
            return False

    def _from_wei(self, wei):
        """Convert wei to ether"""
        return self.server.fromWei(wei, 'ether')

    def get_receive_address(self):
        return self.accounts[0]

    def get_balance(self):
        wei = self.server.eth.getBalance(self.get_receive_address(), 'latest')
        return self._from_wei(wei)

    def suggest_miner_fee(self):
        gas_needed = self.server.eth.estimateGas(
            {
                'from': self.get_receive_address(),
                'to': self.get_receive_address(),
                'value': 1
            }
        )

        # Get the gas price in the latest unconfirmed block
        gas_price = self.server.eth.gasPrice

        return self._from_wei(gas_price * gas_needed)

    def transactions_since(self, since_unix_time, only_look_at=10):
        """
        Gets transactions since specified unix timestamp.
        We use Etherchain.org's server, because it's impractical to create an index for ETH
            transactions ourselves.
        Caveats:
            - this has privacy implications (you reveal your ETH address to Etherchain)
            - only looks at last `only_look_at` transactions to avoid abusing the server
            - only shows 5 decimals after the point (minimum value 0.00001 ETH)
        """


        acc = self.ec.account(self.accounts[0])
        raw_txns = acc.transactions(length=only_look_at, direction='in')
        return self._process_history(raw_txns, since_unix_time, self._timestamp_getter)

    @classmethod
    def _process_history(cls, raw_txns, since_unix_time, timestamp_getter):
        def to_eth(value_string):
            value, unit = value_string.split()
            if unit != 'ETH':
                raise ValueError('Txn not valued in ETH: {}'.format(value_string))
            return Decimal(value)

        txn_list = [
            {
                'txid': tx['parenthash'],
                'time': timestamp_getter(int(tx['blocknumber'])),
                'value': to_eth(tx['value'])
            }
            for tx in raw_txns['data'] if
                tx['direction'].lower() == 'in' and
                tx['value'] != '0 ETH'
        ]

        return [tx for tx in txn_list if tx['value'] > 0 and tx['time'] >= since_unix_time]

    def _timestamp_getter(self, blocknumber):
        if blocknumber not in self._block_timestamps:
            # Try the local Eth server
            logger.debug('Retrieving block #{}...'.format(blocknumber))
            block = self.server.eth.getBlock(blocknumber, full_transactions=False)
            if block:
                self._block_timestamps[blocknumber] = block.timestamp
            else:
                # Fallback to Etherchain.org
                block = self.ec.api.get_block(3865982)
                ts = self._parse_utc_time(block['time'])
                self._block_timestamps[blocknumber] = ts

        return self._block_timestamps[blocknumber]

    @classmethod
    def _parse_utc_time(cls, timestr):
        dt = datetime.datetime.strptime(
                timestr, "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        return calendar.timegm(dt.timetuple())

    def perform_transaction(self, net_amount, miner_fee, target_address):
        """
        Send Ether to target_address (total cost: net_amount + miner_fee)
        """
        assert isinstance(net_amount, Decimal)
        assert isinstance(miner_fee, Decimal)

        net_amount_wei = self.server.toWei(net_amount, 'ether')
        if net_amount_wei != net_amount*Decimal('1e18'):
            raise ValueError('net_amount is not an integer multiple of wei.')

        gas_price_wei = self._get_gas_price(miner_fee)

        txid = self.server.personal.sendTransaction(
            {
                'to': target_address,
                'gas': self.TXN_GAS_LIMIT,
                'gasPrice': gas_price_wei,
                'value': net_amount_wei,
            },
            self.wallet_password
        )

        return txid

    def _get_gas_price(self, eth_to_spend):
        assert isinstance(eth_to_spend, Decimal)

        wei_to_spend = self.server.toWei(eth_to_spend, 'ether');
        gas_price_wei = wei_to_spend/self.TXN_GAS_LIMIT
        if gas_price_wei != int(gas_price_wei):
            raise ValueError('Transaction fee is not an integer multiple of the gas price.')

        return int(gas_price_wei)



import logging
import os
import socket
import subprocess
import pexpect
from pexpect import popen_spawn
from decimal import Decimal
from time import time

import jsonrpclib
import requests

from .processing import inexact_to_decimal, wait_for_success, check_port

logger = logging.getLogger('piggy_logs')

class PiggyXMR:
    # Atomic units in 1 XMR
    ATOMS = Decimal('1e12')

    def __init__(self,
                 daemon_bin_path,
                 wallet_bin_path,
                 datastore_path,
                 wallet_filename,
                 wallet_password,
                 daemon_port,
                 rpcport,
                 ):
        """
        Manage a Monero wallet

        :param daemon_bin_path: Path to Monero daemon executable
        :param wallet_bin_path: Path to Monero RPC wallet executable
        :param datastore_path: Path to datastore directory (which includes wallet files and blockchain)
        :param wallet_filename: Name of wallet file to use (must be in `datastore_path/wallets`)
        :param wallet_password: Password to enter for decrypting wallet
        :param daemon_port: Local port to start Monero daemon on
        :param rpcport: Local port to start the wallet RPC server on
        """

        self.daemon_bin_path = daemon_bin_path
        self.wallet_bin_path = wallet_bin_path
        self.datadir = datastore_path
        self.wallet_filename = wallet_filename
        self.wallet_password = wallet_password

        self.rpcport = check_port(rpcport, 'rpcport')
        self.daemon_port = check_port(daemon_port, 'daemon_port')
        self.daemon_address = '127.0.0.1:{}'.format(self.daemon_port)
        self.server = None

    def start_server(self):
        """Start the daemon and the wallet RPC server"""

        self._check_version()
        self._start_monero_daemon()
        self._start_monero_wallet()

    def _check_version(self):
        """Check that we have the correct version"""
        process = pexpect.spawn(self.wallet_bin_path, ['--version'])

        try:
            process.expect_exact("Monero 'Lithium Luna' (v0.12.2.0-release)")
        except pexpect.exceptions.EOF:
            print('############### OUTPUT ###############')
            print(repr(process.read()))
            print('############### OUTPUT ###############')
            raise ValueError('Error! Version changed on XMR wallet.')

    def _start_monero_daemon(self):
        """Run the Monero daemon"""
        if self._daemon_rpc_loaded():
            logger.info('XMR Daemon already loaded!')
            return

        command = [
            self.daemon_bin_path,

            '--data-dir',
            self.datadir,  # Where to keep data & config

            '--rpc-bind-port={}'.format(self.daemon_port),

            '--detach',
            '--non-interactive',
            '--log-level', '1',
        ]

        logger.info("Starting XMR daemon: {}".format(' '.join(command)))
        subprocess.Popen(command)

        wait_for_success(self._daemon_rpc_loaded, 'Daemon RPC')

    def _daemon_rpc_loaded(self):
        try:
            requests.post("http://{}/getheight".format(self.daemon_address), json={})
            return True
        except requests.exceptions.ConnectionError:
            return False

    def _start_monero_wallet(self):
        """Run the Monero wallet, which connects to the daemon"""

        if self._wallet_rpc_loaded():
            logger.info('XMR Wallet already loaded!')
            return

        command = [
            self.wallet_bin_path,
            '--trusted-daemon',  # We trust our own daemon. If you use a different daemon, remove this
            '--daemon-address',
            self.daemon_address,
            '--rpc-bind-port={}'.format(self.rpcport),
            '--wallet-file',
            os.path.join(self.datadir, 'wallets', self.wallet_filename),
            '--disable-rpc-login',
            '--prompt-for-password'
        ]

        logger.info("Starting XMR wallet: {}".format(' '.join(command)))
        wallet_proc = popen_spawn.PopenSpawn(command)
        logger.info("Sending password")
        wallet_proc.sendline(self.wallet_password)
        logger.info("Sent password")

        wait_for_success(self._wallet_rpc_loaded, 'Wallet RPC')

    def _wallet_rpc_loaded(self):
        try:
            self.get_balance()
            return True
        except socket.error:
            return False

    def stop_server(self):
        """Stop the wallet and the daemon"""
        self._connect_if_needed()

        try:
            self.server.stop_wallet()
            logger.info("stopped wallet")
        except TypeError:
            logger.warn('Not sure if wallet stopped. Received None response!')

        req = requests.post("http://{}/stop_daemon".format(self.daemon_address))

        if req.status_code == 200:
            logger.info("stopped daemon")
        else:
            raise ValueError('Unable to stop daemon!\nSTATUS: {}\nText:{}'.format(req.status, req.text))

    def _connect_if_needed(self):
        if self.server is None:
            self.server = jsonrpclib.Server("http://127.0.0.1:{}/json_rpc".format(self.rpcport))

    def get_balance(self):
        """Return the amount we have in the wallet, confirmed by the blockchain"""
        self._connect_if_needed()

        # Money is unlocked 10 Monero blocks after receiving (~15 minutes)
        return Decimal(self.server.getbalance()['unlocked_balance']) / self.ATOMS

    def get_receive_address(self):
        """Returns a new address where we can receive XMR"""
        self._connect_if_needed()

        result = self.server.make_integrated_address()
        return result['integrated_address']

    def suggest_miner_fee(self):
        """
        Returns miner fee to use, for typical transaction
        Note: Sadly, we can only do this if we have SOME funds, as far as I know.
            Please submit a pull request if you can do it some other way!
        """

        # TODO: Use get_fee_estimate in the daemon

        self._connect_if_needed()

        result = self.server.transfer(
            destinations=[
                {'address': self.get_receive_address(), 'amount': 1}
            ],
            do_not_relay=True
        )

        return Decimal(result['fee']) / self.ATOMS

    def transactions_since(self, since_unix_time):
        """Gets incoming transactions since specified time

        :param since_unix_time: integer or float, unix time
        :return transactions: DataFrame of transactions
            Each row in the DataFrame has amount, txid, and time of each transaction
        """

        self._connect_if_needed()

        history = self.server.get_transfers(**{'in': True})

        return self._process_history(history, since_unix_time)

    @classmethod
    def _process_history(cls, history, since_unix_time):
        """Process the transaction history data"""

        if 'in' in history.keys():
            history_income = history['in']
            return [
                {
                    'txid': txn['txid'],
                    'time': txn['timestamp'],
                    'value': inexact_to_decimal(txn['amount']) / cls.ATOMS,
                }
                for txn in history_income
                if
                    txn['timestamp'] >= float(since_unix_time)
                and
                    txn['type'] == 'in'
                and
                    txn['amount'] > 0
            ]
        else:
            # We don't have the "in" key in the dict, so we don't have any incoming transactions.
            return []

    def perform_transaction(self, net_amount, miner_fee, target_address):
        """
        Send Monero to target_address (total cost: net_amount + miner_fee)

        Note: here, miner_fee is only used as a check (must equal recent `suggest_miner_fee` output).
        """

        self._connect_if_needed()

        assert isinstance(net_amount, Decimal)
        assert isinstance(miner_fee, Decimal)

        # Convert units (Multiply by ATOMS)
        net_amount_atoms = net_amount * self.ATOMS
        miner_fee_atoms = miner_fee * self.ATOMS

        # Check that the amounts are expressible as integer atoms
        assert int(net_amount_atoms) == net_amount_atoms
        assert int(miner_fee_atoms) == miner_fee_atoms

        tx = self.server.transfer(
            destinations=[
                {'address': target_address, 'amount': int(net_amount_atoms)}
            ],
            do_not_relay=True,
            get_tx_hex=True
        )

        if tx['fee'] != miner_fee_atoms:
            raise ValueError(
                'Miner fee no longer equals the desired fee. Please use a freshly estimated fee, and make sure the wallet is up-to-date with the blockchain.\n'
                'Desired fee: {}\n'
                'New estimated fee: {}\n'.format(
                    miner_fee, Decimal(tx['fee']) / self.ATOMS
                )
            )

        success = self._broadcast( tx)

        if success:
            logger.warning("Signed TX: {}".format(tx))
            logger.warning("Broadcast TX ID: {}".format(tx['tx_hash']))
            return tx['tx_hash']
        else:
            raise BaseException("XRP Transaction broadcast fail for: {}".format(tx))

    def _broadcast(self, tx):
        """Broadcast the transaction to the blockchain

        :param tx: Dictionary containing transaction data (tx_blob is mandatory)
        """

        req = requests.post(
            "http://{}/send_raw_transaction".format(self.daemon_address),
            json={'tx_as_hex': tx['tx_blob'], 'do_not_relay': False}
        )

        error = req.json().get('error')
        logger.warning("Broadcasted transaction: {}".format(req.json()))

        if error is not None or req.json()['status'] == 'Failed':
            raise ValueError(
                "Error Broadcasting transaction: ({})\n"
                "TXN: {}\n".format(error, tx)
            )
        elif req.ok:
            return True
        else:
            raise ValueError("Unsuccessful request to broadcast transaction: {}")

    def _refresh(self):
        logger.debug('Rescanning blockchain for wallet...')
        t1 = time()
        res = self.server.rescan_blockchain()
        t2 = time()
        logger.debug("Blockchain Rescanned (took {}s): {}".format(t2-t1, res))

        logger.debug('flushing tx pool for daemon...')
        t1 = time()
        res = requests.post(
            "http://{}/flush_txpool".format(self.daemon_address),
            json={}
        )
        t2 = time()
        logger.debug("Flushed tx pool (took {}s): {}".format(t2-t1, res))
